import os.path
from croydon.baseapp import BaseApp


class App(BaseApp):

    def __init__(self):
        app_dir = os.path.abspath(os.path.dirname(__file__))
        project_dir = os.path.abspath(os.path.join(app_dir, ".."))
        super().__init__(project_dir=project_dir)

    def setup_routes(self) -> None:
        from .controllers.api.v1.account import account_ctrl
        self.include_router(account_ctrl)


app = App()



import logging
import logging.handlers
import os
import sys
from logging import Logger, getLogger
from typing import Generic, TypeVar, Type, Optional
from .types import EnvironmentType, CacheType
from .config import BaseConfig
from .taskq.types import TBaseQueue
from .cache import TCache, NoCache, SimpleCache, MemcachedCache, RequestLocalCache
from .db import DB
from .errors import ConfigurationError

_LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

C = TypeVar("C", bound=BaseConfig)
CT = Type[C]


# To avoid initialization issues Context must be fully synchronous
# i.e. no properties requiring async initialization are allowed.
# Thus, db moved out. It's not going to be needed anywhere besides
# StorableModel anyway.
class Context(Generic[C]):
    _env: Optional[EnvironmentType] = None
    _cfg: Optional[C] = None
    _log: Optional[Logger] = None
    _queue: Optional[TBaseQueue] = None
    _cache_l1: Optional[TCache] = None
    _cache_l2: Optional[TCache] = None
    _db: DB = None

    _project_dir: str
    _initialised = False

    @property
    def project_dir(self):
        return self._project_dir

    @property
    def initialised(self):
        return self._initialised

    @property
    def env(self) -> EnvironmentType:
        if self._env is None:
            raise RuntimeError("attempted to use uninitialised env")
        return self._env

    @property
    def cfg(self) -> C:
        if self._cfg is None:
            raise RuntimeError("attempted to use uninitialised config")
        return self._cfg

    @property
    def log(self) -> Logger:
        if self._log is None:
            raise RuntimeError("attempted to use uninitialised logging")
        return self._log

    @property
    def queue(self) -> TBaseQueue:
        if self._queue is None:
            raise RuntimeError("attempted to use uninitialised queue")
        return self._queue

    @property
    def cache_l1(self) -> TCache:
        if self._cache_l1 is None:
            raise RuntimeError("attempted to use uninitialised cache l1")
        return self._cache_l1

    @property
    def cache_l2(self) -> TCache:
        if self._cache_l2 is None:
            raise RuntimeError("attempted to use uninitialised cache l2")
        return self._cache_l2

    @property
    def db(self) -> DB:
        if self._db is None:
            raise RuntimeError("attempted to use uninitialised DB")
        return self._db

    def setup(self,
              project_dir: str,
              config: CT | C,
              env: Optional[EnvironmentType] = None) -> None:

        self._project_dir = project_dir

        if self._env is None:
            self._setup_env(env)

        if isinstance(config, BaseConfig):
            self._cfg = config
        else:
            self._cfg = self._setup_config(config)

        self._setup_logging()

        self._setup_cache()
        self._setup_db()  # needs logging and cache
        self._setup_queue()  # needs db
        self._initialised = True

    def _setup_db(self):
        self._db = DB(
            config=self._cfg.database,
            mock=False,
        )

    def _setup_cache(self):
        for prop in ["level1", "level2"]:
            cache_type: CacheType = getattr(self._cfg.cache, prop)

            if cache_type == "none":
                cache = NoCache()
            elif cache_type == "simple":
                cache = SimpleCache()
            elif cache_type == "memcached":
                backends = self._cfg.memcached.backends
                cache = MemcachedCache(backends)
            elif cache_type == "request_local":
                cache = RequestLocalCache()
            else:
                raise ConfigurationError(f"invalid cache type {cache_type}")

            if prop == "level1":
                self._cache_l1 = cache
            else:
                self._cache_l2 = cache

    def _setup_queue(self):
        if self._cfg.queue.type == "mongo":
            from .taskq.mongo_queue import MongoQueue
            self._queue = MongoQueue(self._cfg.mongo_queue)
        else:
            raise TypeError(f"queue type {self._cfg.queue.type} is invalid")

    def _setup_config(self, cfgcls: CT) -> C:
        config_filename = os.path.join(self._project_dir, f"{self._env}.toml")
        return cfgcls.parse(config_filename)

    def _setup_env(self, env: EnvironmentType = None):
        if env is None:
            env = "development"
            ext_env = os.getenv("CROYDON_ENV")
            if ext_env in ["development", "testing", "staging", "production"]:
                env = ext_env
        self._env = env

    def _setup_logging(self):
        logger = getLogger(__name__)
        logger.propagate = False

        lvl = _LOG_LEVELS.get(
            self._cfg.logging.level.lower(),
            logging.DEBUG
        )
        logger.setLevel(lvl)

        for handler in logger.handlers:
            logger.removeHandler(handler)

        log_format = logging.Formatter(
            "[%(asctime)s] %(levelname)s %(filename)s:%(lineno)d %(message)s"
        )

        if self._cfg.logging.stdout:
            handler = logging.StreamHandler(stream=sys.stdout)
            handler.setLevel(lvl)
            handler.setFormatter(log_format)
            logger.addHandler(handler)

        if self._cfg.logging.filename is not None:
            handler = logging.handlers.WatchedFileHandler(self.cfg.logging.filename)
            handler.setLevel(lvl)
            handler.setFormatter(log_format)
            logger.addHandler(handler)

        self._log = logger


ctx = Context()



from typing import Optional, Callable, Awaitable
from fastapi import Header, Depends, Request
from croydon.errors import AuthenticationError
from app.models import Token, User, Session


async def auth_token(authorization: Optional[str] = Header(default=None)) -> Optional[Token]:
    if authorization is None:
        return None

    parts = authorization.split(" ")
    if len(parts) != 2:
        return None
    token_type, token = parts
    if token_type != "Token":
        return None

    return await Token.cache_get(token)


def authenticated_user(required: bool = True) -> Callable[[Optional[Token]], Awaitable[Optional[User]]]:
    async def inner(token: Optional[Token] = Depends(auth_token)) -> Optional[User]:
        user: Optional[User] = None
        if token is not None:
            user = await token.user()
        if user is None and required:
            raise AuthenticationError()
        return user
    return inner


async def current_session(request: Request) -> Session:
    return request.state.session



from typing import TypeVar, TYPE_CHECKING, Type, TypedDict, Literal
from datetime import datetime
if TYPE_CHECKING:
    from .base_queue import BaseQueue
    from .task import BaseTask


TBaseQueue = TypeVar("TBaseQueue", bound="BaseQueue")
TBaseTask = TypeVar("TBaseTask", bound="BaseTask")
TBaseTaskType = Type[TBaseTask]
TTaskData = TypeVar("TTaskData")

TState = Literal["new", "published", "picked_up", "done", "error"]


class Message(TypedDict):
    id: str
    type: str
    data: str
    created_at: datetime
    state: TState
    status_message: str



from croydon.taskq.worker import BaseWorker
from croydon.taskq.types import TBaseTask


class Worker(BaseWorker):

    async def run_task(self, task: TBaseTask) -> None:
        raise NotImplemented(
            "extend Worker class in app.tasks.worker to teach it "
            "how to process tasks of various types"
        )



import os
import importlib
import inspect
from typing import Dict, Type
from croydon import ctx
from croydon.models.storable_model import StorableModel
from croydon.models.counter import Counter
from croydon.command import Command


class Index(Command):

    NAME = "index"
    HELP = "Create indexes for the application models"

    async def run_async(self) -> None:
        from app import app
        app.initialise()

        models_dir = os.path.join(ctx.project_dir, "app/models")
        classes: Dict[str, Type['StorableModel']] = {
            "Counter": Counter
        }

        for filename in os.listdir(models_dir):
            if filename.endswith(".py") and not filename.startswith("__"):
                modulename = filename[:-3]
                module = importlib.import_module(f"app.models.{modulename}")

                for attr in dir(module):
                    cls = getattr(module, attr)
                    if inspect.isclass(cls) and issubclass(cls, StorableModel) and not cls is StorableModel:
                        classes[cls.__name__] = cls

        for cls_name, cls in classes.items():
            ctx.log.debug(f"Creating index for model {cls_name}")
            await cls.ensure_indexes(loud=True)



from typing import Optional
from fastapi import HTTPException


class ApiError(HTTPException):

    error_key: str = "api_error"
    status_code: int = 500

    def __init__(self, message: Optional[str] = None):
        super(ApiError, self).__init__(status_code=self.status_code, detail=message)


class AuthenticationError(ApiError):
    status_code = 401
    error_key = "auth_error"

    def __init__(self, message: str = "you must be authenticated first") -> None:
        super().__init__(message)


class ConfigurationError(SystemExit):
    error_key = "configuration_error"


class Forbidden(ApiError):
    error_key = "forbidden"
    status_code = 403


class BadRequest(ApiError):
    error_key = "bad_request"
    status_code = 400


class Conflict(ApiError):
    error_key = "conflict"
    status_code = 409


class IntegrityError(ApiError):
    error_key = "integrity_error"
    status_code = 409


class NotFound(ApiError):
    error_key = "not_found"
    status_code = 404


class InternalServerError(ApiError):
    error_key = "intrnl_error"
    status_code = 500


class InvalidShardId(InternalServerError):
    error_key = "intrnl_error"


class ShardIsReadOnly(IntegrityError):
    error_key = "shard_readonly"


class ModelDestroyed(IntegrityError):
    error_key = "model_destroyed"


class InputDataError(BadRequest):
    error_key = "bad_input"


class InvalidFieldType(ApiError):
    error_key = "bad_input_type"


class OAuthError(ApiError):
    error_key = "oauth_error"
    status_code = 401


class ORMException(Exception):
    pass


class AbortTransaction(ORMException):
    def __init__(self) -> None:
        super().__init__("transaction aborted")


class ValidationError(ORMException, ApiError):
    error_key = "validation_error"
    status_code = 422


class DoNotSave(ORMException):
    pass


class ObjectSaveRequired(ORMException):
    pass


class ObjectHasReferences(ORMException, ApiError):
    status_code = 409
    error_key = "object_has_references"


class MissingSubmodel(ORMException):
    pass


class WrongSubmodel(ORMException):
    pass


class UnknownSubmodel(ORMException):
    pass



from abc import ABC
from ..models.storable_model import StorableModel
from ..models.meta_model import ModelReference
from ..models.fields import ReferenceField, StringField, ReferenceOnDestroy
from ..errors import ValidationError, ObjectHasReferences
from .mongo_mock_test import MongoMockTest


class TestReference(MongoMockTest):

    async def test_reference_field_raise(self):
        class Master(StorableModel):
            KEY_FIELD = "name"
            name = StringField()

        class Minion(StorableModel):
            KEY_FIELD = "name"
            name = StringField()
            master_id: ReferenceField[Master] = ReferenceField(reference_model=Master, required=True)

        m = Master.create(name="master")
        await m.save()

        mi1 = Minion.create(name="mi1", master_id="abc")

        with self.assertRaises(ValidationError) as cx:
            await mi1.save()

        mi1.master_id = m.id
        await mi1.save()

        self.assertEqual("Broken reference master_id: no Master found", cx.exception.detail)

        with self.assertRaises(ObjectHasReferences) as cx:
            await m.destroy()

    async def test_reference_field_cascade(self):
        class Master(StorableModel):
            KEY_FIELD = "name"
            name = StringField()

        class Minion(StorableModel):
            KEY_FIELD = "name"
            name = StringField()
            master_id: ReferenceField[Master] = ReferenceField(reference_model=Master, required=True,
                                                               on_destroy=ReferenceOnDestroy.CASCADE)

        m = Master.create(name="master")
        await m.save()
        mi1 = Minion.create(name="mi1", master_id=m.id)
        await mi1.save()
        self.assertEqual(1, await Minion.find({}).count())

        await m.destroy()
        self.assertEqual(0, await Minion.find({}).count())

    async def test_references_to_subclasses(self):
        class Master(StorableModel):
            KEY_FIELD = "name"
            name = StringField()

        class Minion(StorableModel):
            KEY_FIELD = "name"
            name = StringField()
            master_id: ReferenceField[Master] = ReferenceField(reference_model=Master, required=True)

        class SubMinion(Minion):
            other_field = StringField()

        self.assertCountEqual(
            [
                ModelReference(ref_class=Minion, ref_field="master_id"),
                ModelReference(ref_class=SubMinion, ref_field="master_id")
            ],
            Master._references
        )

    async def test_references_to_subclasses_of_abc(self):
        class Master(StorableModel):
            KEY_FIELD = "name"
            name = StringField()

        class Minion(StorableModel, ABC):
            KEY_FIELD = "name"
            name = StringField()
            master_id: ReferenceField[Master] = ReferenceField(reference_model=Master, required=True)

        class SubMinion(Minion):
            other_field = StringField()

        self.assertCountEqual(
            [ModelReference(ref_class=SubMinion, ref_field="master_id")],
            Master._references
        )



from typing import Dict, Any, Optional
from unittest import IsolatedAsyncioTestCase
from contextvars import ContextVar, Token
from ..context import ctx
from ..config import DatabaseConfig
from ..cache import TraceCache, RequestLocalCache
from ..middleware import REQUEST_CACHE_CONTEXT_KEY


custom_cache_ctxvar: ContextVar[Optional[Dict[str, Any]]] = ContextVar(REQUEST_CACHE_CONTEXT_KEY, default=None)


class MongoMockTest(IsolatedAsyncioTestCase):

    token: Optional[Token[Dict[str, Any]]] = None

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        if not ctx.initialised:
            from ..baseapp import BaseApp
            app = BaseApp(".")
            app.initialise()
        ctx.db.reconfigure(DatabaseConfig(), mock=True)

        # aiomcache conflicts with async tests as it seems to run
        # in its own async loop. This leads to "got Future attached to a different loop"
        # errors.
        #
        # Since mongo tests do not use memcached explicitly we override
        # all caches to avoid using MemcachedCache
        #
        # RequestLocalCache is created using a custom contextvar so that
        # L2 Cache resets after each test rather than after each request
        #
        # TraceCache does not store any data but is able to track its methods calls
        ctx._cache_l1 = TraceCache()
        ctx._cache_l2 = RequestLocalCache(custom_cache_ctxvar)

    async def asyncSetUp(self) -> None:
        self.token = custom_cache_ctxvar.set({})

    async def asyncTearDown(self) -> None:
        custom_cache_ctxvar.reset(self.token)
        self.token = None


